/*****************************************************************************
 * Copyright (C) The Apache Software Foundation. All rights reserved.        *
 * ------------------------------------------------------------------------- *
 * This software is published under the terms of the Apache Software License *
 * version 1.1, a copy of which has been included with this distribution in  *
 * the LICENSE file.                                                         *
 *****************************************************************************/

package org.apache.batik.bridge;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import org.apache.batik.css.AbstractViewCSS;
import org.apache.batik.css.CSSOMReadOnlyStyleDeclaration;
import org.apache.batik.css.CSSOMReadOnlyValue;
import org.apache.batik.dom.svg.SVGOMDocument;
import org.apache.batik.dom.util.XLinkSupport;
import org.apache.batik.util.ParsedURL;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.svg.SVGDocument;

/**
 * This class is used to resolve the URI that can be found in a SVG document.
 *
 * @author <a href="mailto:stephane@hillion.org">Stephane Hillion</a>
 * @version $Id: URIResolver.java,v 1.8 2001/07/23 19:36:29 deweese Exp $
 */
public class URIResolver {
    /**
     * The reference document.
     */
    protected SVGOMDocument document;

    /**
     * The document URI.
     */
    protected String documentURI;

    /**
     * The document loader.
     */
    protected DocumentLoader documentLoader;

    /**
     * Creates a new URI resolver object.
     * @param doc The reference document.
     * @param dl The document loader.
     */
    public URIResolver(SVGDocument doc, DocumentLoader dl) {
        document = (SVGOMDocument)doc;
        documentLoader = dl;
    }

    /**
     * Imports the element referenced by the given URI.
     * @param uri The element URI.
     */
    public Element getElement(String uri)
        throws MalformedURLException, IOException {
        Node n = getNode(uri);
        if (n == null) {
            return null;
        } else if (n.getNodeType() == n.DOCUMENT_NODE) {
            throw new IllegalArgumentException();
        } else {
            return (Element)n;
        }
    }

    /**
     * Returns the node referenced by the given URI.
     * @return The document or the element
     */
    public Node getNode(String uri)
        throws MalformedURLException, IOException {

        if (uri.startsWith("#")) {
            return document.getElementById(uri.substring(1));
        }

        if (documentURI == null)
            documentURI = document.getURL();

        ParsedURL purl = new ParsedURL(documentURI, uri);
        String    ref  = purl.getRef();
        if ((ref != null) && (documentURI != null)) {
            ParsedURL pDocURL = new ParsedURL(documentURI);
            // Check if the rest of the URL matches...
            // if so then return the referenced element.
            if ((pDocURL.getPath() == purl.getPath()) &&
                (pDocURL.getPort() == purl.getPort()) &&
                (pDocURL.getHost() == purl.getHost()) &&
                (pDocURL.getProtocol() == purl.getProtocol()))
                return document.getElementById(ref);
        }

        // uri is not a reference into this document, so load the 
        // document it does reference.
        Document doc = documentLoader.loadDocument(purl.toString());
        if (ref != null)
            return doc.getElementById(ref);
        return doc;
    }
}
